-- This work is licensed under a Creative Commons
-- Attribution-NonCommercial-ShareAlike 4.0 International License.
-- http://creativecommons.org/licenses/by-nc-sa/4.0

-- 2020 modifications:  Mark D. Aagaard, Nusa Zidaric

-- Mark D. Aagaard
-- Riham AlTawy
-- Guang Gong
-- Kalikinkar Mandal
-- Raghvendra Rohit
-- Marat Sattarov
-- Nusa Zidaric
-- http://uwaterloo.ca/communications-security-lab/lwc/wage


-- This is a human-readable summary of (and not a substitute for) the license. 
-- You are free to:

--     Share:  copy and redistribute the material in any medium or format
--     Adapt:  remix, transform, and build upon the material

--     The licensor cannot revoke these freedoms as long as you follow
--     the license terms.

-- Under the following terms:

--     Attribution - You must give appropriate credit, provide a link to
--     the license, and indicate if changes were made. You may do so in
--     any reasonable manner, but not in any way that suggests the
--     licensor endorses you or your use.

--     NonCommercial - You may not use the material for commercial
--     purposes.

--     ShareAlike - If you remix, transform, or build upon the material,
--     you must distribute your contributions under the same license as
--     the original.

--     No additional restrictions - You may not apply legal terms or
--     technological measures that legally restrict others from doing
--     anything the license permits.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.wage_pkg.all;
use work.util.all;
use work.ctrl_pkg.all;
  
entity ctrl is
  port
    ( clk
    , reset         	: in  std_logic
    ; pre_mode       : in  mode_ty
    ; pre_dom_sep    : in  dom_sep_ty
    ; pre_valid     	: in  std_logic
    ; post_ready     : in  std_logic
    ; pre_padding    : in  full_pad_ty
    ; o_control     	: out dp_ctl_ty
    ; o_ce_ctrl 	  	: out ce_ty  
    ; o_pad_ctrl	  	: out padxor_ty
    ; post_valid     : out std_logic
    ; pre_ready      : out std_logic

   );
end entity;
 
architecture rtl of ctrl is
  
  signal state,  state_next	: state_ty;
  signal count  					: count_ty;
  signal k_count  				: std_logic;
   
  signal ce_ctrl 	    : ce_ty;
  signal pad_ctrl 	 : padxor_ty;
  signal control 		 : dp_op_ty ;

  signal exit_l_load, exit_p_load, exit_p_init
  		,  exit_p_aed, exit_p_final, exit_t_tag : 		std_logic;
  		
  signal rollover, counting, flip, fullpad
  		, 	valid_ed, valid_t : 							std_logic;
  		
  signal q_load, q_init, q_aedf, q_ed, q_tag  
		,  r_load, r_init, r_aedf, r_ed, r_af  						 
		,	ds_aedf, ds_af, z_tag
		,	v_ed, v_tag:  						std_logic; 	
  
signal   condition_next_i_load,
    		condition_next_l_load ,
    		condition_next_p_load ,
    		condition_next_i_init ,  
    		condition_next_p_init ,
    		condition_next_i_aedf ,
	 		condition_next_p_aed ,
    		condition_next_p_final,
    		condition_next_t_tag  ,
    		condition_next_s_tag 	: std_logic;									
     
  begin

  ------------------------------------------------------------
  -- environment control   
  ------------------------------------------------------------
  pre_ready 	<= '1'  when   x2b(  r_load or r_init or r_af or r_ed ) else 
  						'0';
  post_valid 	<= valid_ed when x2b( v_ed ) 	else 
  						valid_t;
  valid_ed 		<= not fullpad ; 
  valid_t  		<= '1'  when x2b(v_tag) 	else 
  						'0';
     						
  ------------------------------------------------------------
  -- datapath control   
  ------------------------------------------------------------

  o_control(load_idx) 				<= 	control(load_idx);
  o_control(sb_off_idx) 			<= 	control(sb_off_idx);
  o_control(lfsr_en_idx)			<= 	control(lfsr_en_idx); 								 																			 
  o_control(absorb_idx) 			<= 	control(absorb_idx); 
  o_control(replace_idx) 			<= 	control(replace_idx); 
  o_control(lfsr_c_reset_idx) 	<= 	(is_perm_done(count) or reset) when x2b(is_perm_st(state))	else
  						 					 		(is_lt_done(count)   or reset) when x2b(is_lt_st(state))    else 
  						  							'0';    						 

  u_ce : entity work.ctrl_pad_to_ce port map (  pre_padding(2 downto 0) ,  ce_ctrl ,   pad_ctrl  );
 
  fullpad 		<= pre_padding(3);
  o_ce_ctrl 	<= ce_ctrl 	when  x2b( v_ed and is_dec_mode(pre_mode) and not fullpad )  else 
  						(others => '1'); 
  o_pad_ctrl 	<= pad_ctrl when  x2b( v_ed and is_dec_mode(pre_mode) and not fullpad )  else 
  						(others => '0');  
  
  control 	<= 	dp_op_1 		when x2b( q_load  and pre_valid)  										else				
  						dp_op_2 		when x2b( is_perm_st(state) and not is_perm_done(count))			else 
  						dp_op_3 		when x2b((q_init and pre_valid)		or		ds_af)					else
  						dp_op_4 		when x2b( v_ed and is_enc_mode(pre_mode) )							else
 						dp_op_5_fp	when x2b( v_ed and is_dec_mode(pre_mode) and fullpad  )  	 	else
 						dp_op_5		when x2b( v_ed and is_dec_mode(pre_mode) and not fullpad )  	else
						dp_op_6 		when x2b( v_tag	) 															else
						(others => '0');							          														  						

  ------------------------------------------------------------
  -- transient states
  ------------------------------------------------------------
  
	  exit_l_load  <= is_l_load_st(state ) and is_lt_done( count );
	  exit_p_load 	<= is_p_load_st(state)  and is_perm_done( count );
	  exit_p_init  <= is_p_init_st(state) 	and is_perm_done( count );
	  exit_p_aed  	<= is_p_aed_st(state) 	and is_perm_done( count );
	  exit_p_final <= is_p_final_st(state) and is_perm_done( count ); 
	  exit_t_tag  	<= is_t_tag_st(state) 	and is_lt_done( count );   
	
    r_load <= is_i_load_st( state ) or											-- from idle
            ( is_l_load_st( state ) and not is_lt_done( count ) );   -- loop back

    q_load <= r_load;
	 
    r_init <= is_i_init_st( state ) 		or 								-- from idle
			   ( exit_p_load )        			or 								-- first in 
  			   ( exit_p_init and  k_count );   								 	-- loop back
  			   
    q_init <= r_init;

    q_aedf <= 	  is_i_aedf_st( state ) 				or 					-- from idle/stall     	
			  		( exit_p_init  and not k_count )  	or						-- first in 
           	 	( exit_p_aed )  							or	 					-- loop back 1 (from pAED)  
            	( exit_p_final and 	  k_count );							-- loop back 2 (from pFinal)
            
    ds_aedf <= q_aedf and pre_valid;
    q_ed  	<= ds_aedf and     is_ed_ds( pre_dom_sep );
    r_af  	<= ds_aedf and not is_ed_ds( pre_dom_sep );  													 

    r_ed  	<= q_ed and post_ready;
    v_ed 	<= r_ed;
    ds_af  	<= r_af;
    
    q_tag <= is_i_tag_st( state ) or											-- from idle 
    			(exit_p_final and not k_count) or 								-- first in 
    			(is_t_tag_st(state) and not is_lt_done(count));			   -- loop back 		 
    			
    v_tag <= q_tag and post_ready;
    
  ------------------------------------------------------------
  -- next state logic
  ------------------------------------------------------------
				
    counting <= 	'1' when x2b(( q_load and pre_valid ) 	or 
    										is_perm_st(state) 		or 
    										is_t_tag_st(state)  ) 	else
    					'0';
    					 
    rollover <=  ( is_lt_st(state) 		and is_lt_done(count)) 	or 
    				  ( is_perm_st (state) 	and is_perm_done( count ))  ;
    				    
     flip 	 <=  (q_init and pre_valid )  	or 
     				  (ds_aedf and is_if_ds( pre_dom_sep ));

    condition_next_l_load 	<=   q_load and pre_valid; 											-- wait for parcel
    condition_next_p_load 	<=   exit_l_load or														-- start perm
							  			( is_p_load_st(state) and not is_perm_done(count));   	-- continue perm   
    condition_next_i_init 	<=   q_init and not pre_valid;	    								-- wait for parcel
    condition_next_p_init 	<= ( q_init and pre_valid ) or 										-- start perm
    									( is_p_init_st(state) and not is_perm_done(count));   	-- continue perm    
    condition_next_i_aedf 	<= ( q_aedf and not pre_valid )  or 								-- wait for parcel
    									( q_ed   and not post_ready ) or 								-- stall	
										( is_i_aedf_st(state) and not  post_ready)    ;				-- still stalling
	 condition_next_p_aed  	<= ( ds_aedf and is_ad_ds( pre_dom_sep ))   or					-- start perm: parcel is AD
	 							  		  v_ed  or 																-- start perm: parcel is ptxt/ctxt									 
	  									( is_p_aed_st(state) and not is_perm_done( count )) ; 	-- continue perm
    condition_next_p_final	<=	( ds_aedf and is_if_ds( pre_dom_sep ) ) or 					-- start perm: parcel is K0/K1
    									( is_p_final_st(state) and not is_perm_done( count ));	-- continue perm
    condition_next_t_tag  	<= q_tag and post_ready;												-- next tag chunk
    condition_next_s_tag 	<= q_tag and not post_ready;											-- stall										
 
           state_next      <= l_load_st  when x2b( condition_next_l_load  ) 	else
             						p_load_st  when x2b( condition_next_p_load  ) 	else
             						i_init_st  when x2b( condition_next_i_init  )	else
             						p_init_st  when x2b( condition_next_p_init  )	else
             						i_aedf_st  when x2b( condition_next_i_aedf  )	else
             						p_aed_st   when x2b( condition_next_p_aed   )	else
             						p_final_st when x2b( condition_next_p_final )	else
             						t_tag_st   when x2b( condition_next_t_tag   )	else
             						i_tag_st   when x2b( condition_next_s_tag   ) 	else			
 									   i_load_st; 
    
  ------------------------------------------------------------
  -- state update
  ------------------------------------------------------------
    
  state_update: process(clk) begin
    if rising_edge(clk) then
      if x2b( reset ) then
        state  	<= i_load_st;
       else
        state  	<= state_next;
       end if;
    end if;   
  end process;

  count_update: process(clk) begin
    if rising_edge(clk) then
       if x2b( reset ) then
       	count  	<= (others => '0');
       else
       	if 	x2b(rollover) then 
       		count <= (others => '0');
       	elsif x2b(counting ) then 
      		count <= count + 1; 	
        	end if;
       end if;
    end if;   
  end process;
 
  kcount_update: process(clk) begin
    if rising_edge(clk) then
      if x2b( reset ) then
        k_count <= '0';
      elsif x2b(flip) then
        k_count <= not k_count;
      end if;
    end if;  
  end process;
  
end architecture;  
