-- This work is licensed under a Creative Commons
-- Attribution-NonCommercial-ShareAlike 4.0 International License.
-- http://creativecommons.org/licenses/by-nc-sa/4.0

-- 2020 modifications:  Mark D. Aagaard, Nusa Zidaric

-- Mark D. Aagaard
-- Riham AlTawy
-- Guang Gong
-- Kalikinkar Mandal
-- Raghvendra Rohit
-- Marat Sattarov
-- Nusa Zidaric
-- http://uwaterloo.ca/communications-security-lab/lwc/wage


-- This is a human-readable summary of (and not a substitute for) the license. 
-- You are free to:

--     Share:  copy and redistribute the material in any medium or format
--     Adapt:  remix, transform, and build upon the material

--     The licensor cannot revoke these freedoms as long as you follow
--     the license terms.

-- Under the following terms:

--     Attribution - You must give appropriate credit, provide a link to
--     the license, and indicate if changes were made. You may do so in
--     any reasonable manner, but not in any way that suggests the
--     licensor endorses you or your use.

--     NonCommercial - You may not use the material for commercial
--     purposes.

--     ShareAlike - If you remix, transform, or build upon the material,
--     you must distribute your contributions under the same license as
--     the original.

--     No additional restrictions - You may not apply legal terms or
--     technological measures that legally restrict others from doing
--     anything the license permits.


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.util.all;

package ctrl_pkg is

 ------------------------------------------------------------
  subtype dom_sep_ty is std_logic_vector( 1 downto 0 );   -- to prevent mix up with downto/to
 
  function is_if_ds		( dom_sep : dom_sep_ty ) 	return std_logic;  -- init/fin
  function is_ed_ds		( dom_sep : dom_sep_ty ) 	return std_logic;  -- enc/dec
  function is_ad_ds 	   ( dom_sep : dom_sep_ty ) 	return std_logic;  -- proc. ad
    
  subtype mode_ty is std_logic;     -- top lvl input
  function is_enc_mode	( mode : mode_ty ) 	return std_logic;
  function is_dec_mode	( mode : mode_ty ) 	return std_logic;
   
  subtype state_ty is std_logic_vector( 0 to 3 );
 	constant i_load_st : state_ty :=  "0000";  
	constant i_init_st : state_ty :=  "0001";
	constant i_aedf_st : state_ty :=  "0011";
	constant i_tag_st  : state_ty :=  "0010";
	constant p_load_st : state_ty :=  "0100";
	constant p_init_st : state_ty :=  "0101";
	constant p_aed_st  : state_ty :=  "0111";
	constant p_final_st: state_ty :=  "0110";
	constant l_load_st : state_ty :=  "1000";
	constant t_tag_st  : state_ty :=  "1010";
  
 ------------------------------------------------------------
  -- state predicates

  function is_i_load_st 	( s : state_ty ) return std_logic;
  function is_i_init_st 	( s : state_ty ) return std_logic;
  function is_i_aedf_st 	( s : state_ty ) return std_logic;
  function is_p_load_st 	( s : state_ty ) return std_logic;
  function is_p_init_st 	( s : state_ty ) return std_logic;
  function is_p_aed_st   	( s : state_ty ) return std_logic;
  function is_p_final_st 	( s : state_ty ) return std_logic;
  function is_l_load_st		( s : state_ty ) return std_logic;
  function is_t_tag_st		( s : state_ty ) return std_logic;
  function is_i_tag_st   	( s : state_ty ) return std_logic;
 
  ------------------------------------------------------------ 
    
  function is_lt_st      	( s : state_ty ) return std_logic;
  function is_perm_st      ( s : state_ty ) return std_logic;
  
  ------------------------------------------------------------
  -- functions to allow parallel testing of conditions

  function "and" ( guard : std_logic; state : state_ty ) return state_ty;
  function "or"  ( state1, state2 : state_ty ) return state_ty;
 
 ------------------------------------------------------------
  -- count
  ------------------------------------------------------------
  
  subtype count_ty is unsigned( 6 downto 0 );
   
  function is_lt_done( count : count_ty ) return std_logic;
  function is_perm_done( count : count_ty ) return std_logic;
 
 ------------------------------------------------------------
  -- datapath operations
  ------------------------------------------------------------

  subtype  dp_ctl_ty is std_logic_vector( 5 downto 0 );
  subtype  dp_op_ty  is std_logic_vector( 4 downto 0 );

 -- for wagelfsr
  constant load_idx          			: natural := 0;
  constant sb_off_idx        			: natural := 1;
  constant lfsr_en_idx       			: natural := 2;
  constant absorb_idx        			: natural := 3;
  constant replace_idx       			: natural := 4;
  constant lfsr_c_reset_idx			: natural := 5;

  -- load
  constant dp_op_1 		: dp_op_ty := (load_idx      => '1',
                                  		lfsr_en_idx   => '1',
                                 		sb_off_idx    => '1',
                                 		others        => '0');

  -- permutation
  constant dp_op_2 		: dp_op_ty := (lfsr_en_idx   => '1',
                                  		others        => '0');

  -- init, proc_ad, final
  constant dp_op_3 		: dp_op_ty := (absorb_idx    => '1',
                                  		others        => '0');

  -- encrypt
  constant dp_op_4 		: dp_op_ty := ( absorb_idx   			=> '1',
                        		          others        		=> '0');

  -- decrypt
  constant dp_op_5 		: dp_op_ty := ( absorb_idx    		=> '1',
  													 replace_idx   		=> '1',
                        		          others        		=> '0');
  -- decrypt full pad
  constant dp_op_5_fp 	: dp_op_ty := ( absorb_idx    => '1',
                        		          others        => '0');
  -- tag
  constant dp_op_6 		: dp_op_ty := (lfsr_en_idx   => '1',
                                 		sb_off_idx    => '1',
                                  		others        => '0');
end package;

 

package body ctrl_pkg is
  ------------------------------------------------------------
  -- dom sep and mode functions
  
  -- init/fin  
  function is_if_ds( dom_sep : dom_sep_ty ) return std_logic is
  begin
    return b2x( dom_sep = "00" );
  end function;
  
  -- enc/dec
  function is_ed_ds		( dom_sep : dom_sep_ty ) 	return std_logic is
  begin
    return b2x( dom_sep(1)  = '1' );
  end function;
  
  -- proc. ad 
  function is_ad_ds		( dom_sep : dom_sep_ty ) 	return std_logic is
  begin
    return b2x( dom_sep(0)  = '1' );
  end function; 
  
   
  function is_enc_mode	( mode : mode_ty ) 	return std_logic is
  begin
    return b2x( mode = '0' );
  end function; 
  
  
  function is_dec_mode	( mode : mode_ty ) 	return std_logic is
  begin
    return b2x( mode = '1' );
  end function;    
   
   
  ------------------------------------------------------------
  -- state predicates

  function is_i_load_st( s : state_ty ) return std_logic is
  begin
    return b2x( s = i_load_st); 
  end function;
  
  function is_i_init_st( s : state_ty ) return std_logic is
  begin
    return b2x( s = i_init_st); 
  end function;
  
  function is_i_aedf_st( s : state_ty ) return std_logic is
  begin
    return b2x( s = i_aedf_st); 
  end function;
  
  function is_p_load_st( s : state_ty ) return std_logic is
  begin
    return b2x( s = p_load_st); 
  end function;
  
  function is_p_init_st( s : state_ty ) return std_logic is
  begin
    return b2x( s = p_init_st);
  end function;
  
  function is_p_aed_st( s : state_ty ) return std_logic is
  begin
    return b2x( s = p_aed_st);
  end function;
  
  function is_p_final_st( s : state_ty ) return std_logic is
  begin
    return b2x( s = p_final_st);
  end function;

  function is_l_load_st( s : state_ty ) return std_logic is
  begin
    return b2x( s = l_load_st);
  end function;
 
  function is_t_tag_st( s : state_ty ) return std_logic is
  begin
    return b2x( s = t_tag_st);
  end function;

  function is_i_tag_st( s : state_ty ) return std_logic is
  begin
    return b2x( s = i_tag_st);
  end function;
  
  
  ------------------------------------------------------------

  function is_lt_st( s : state_ty ) return std_logic is
  begin	
    return b2x( s(0) = '1');
  end function;
    
  function is_perm_st( s : state_ty ) return std_logic is
  begin	
    return b2x( s(1) = '1');
  end function;
 
  
  ------------------------------------------------------------
  -- functions to allow parallel testing of conditions

  function "and"( guard : std_logic; state : state_ty ) return state_ty is
  begin
    if x2b( guard ) then
      return state;
    else
      return (others => '0');
    end if;
  end function;

  function "or"( state1, state2 : state_ty ) return state_ty is
    variable res : state_ty;
  begin
    for i in state1'range loop
      res(i) := state1(i) or state2(i);
    end loop;
    return res;
  end function;


  --------------------------------------------------------------
  -- unified is_load_done and is_tag_done: 
  function is_lt_done( count : count_ty ) return std_logic is
  begin
  return b2x(count = 9);
  end function;

  
 
  function is_perm_done( count : count_ty ) return std_logic is
  begin
  return b2x(count = 111);
  end function;
 
  

end package body;  





