-- This work is licensed under a Creative Commons
-- Attribution-NonCommercial-ShareAlike 4.0 International License.
-- http://creativecommons.org/licenses/by-nc-sa/4.0

-- 2020 modifications:  Mark D. Aagaard, Nusa Zidaric

-- Mark D. Aagaard
-- Riham AlTawy
-- Guang Gong
-- Kalikinkar Mandal
-- Raghvendra Rohit
-- Marat Sattarov
-- Nusa Zidaric
-- http://uwaterloo.ca/communications-security-lab/lwc/wage


-- This is a human-readable summary of (and not a substitute for) the license. 
-- You are free to:

--     Share:  copy and redistribute the material in any medium or format
--     Adapt:  remix, transform, and build upon the material

--     The licensor cannot revoke these freedoms as long as you follow
--     the license terms.

-- Under the following terms:

--     Attribution - You must give appropriate credit, provide a link to
--     the license, and indicate if changes were made. You may do so in
--     any reasonable manner, but not in any way that suggests the
--     licensor endorses you or your use.

--     NonCommercial - You may not use the material for commercial
--     purposes.

--     ShareAlike - If you remix, transform, or build upon the material,
--     you must distribute your contributions under the same license as
--     the original.

--     No additional restrictions - You may not apply legal terms or
--     technological measures that legally restrict others from doing
--     anything the license permits.

library ieee;
use ieee.std_logic_1164.all;
use work.wage_pkg.all;
use work.ctrl_pkg.all;

-- 4 types of MUX and AND-mask for tag
-- CE, MUX, NOT for padding positions (DEC padding control)
 
entity lfsr_wage is
  port
    ( clk 						 : in  std_logic
    ; ce_ctrl 					 : in  ce_ty
    ; pad_ctrl   				 : in  padxor_ty
    ; ctrl						 : in  dp_op_ty
    ; i_dom_sep             : in  dom_sep_ty
    ; i_data                : in  data_io
    ; i_s36, i_s29, i_s23   : in  gf_elem --  left half nonlinears (inputs)
    ; i_s18, i_s10, i_s4    : in  gf_elem -- right half nonlinears (inputs)  
    ; o_s36, o_s34, o_s27   : out gf_elem --  left half nonlinears (outputs)
    ; o_s18, o_s15, o_s8    : out gf_elem -- right half nonlinears (outputs)
    ; o_data                : out data_io
   );
end entity;

architecture rtl of lfsr_wage is
  signal f, omega : gf_elem;  -- feedback
  signal i_data_9 : gf_elem;  -- for loading
  signal sa : state_array;    -- wage state 
  signal x,                   -- result of XORing with nonlinears
         ab,                  -- result of absorbing  
         y,                   -- choose between ab and sa/x 
         l  : state_array;    -- choose betweeb y and i_data
    
 
  signal replace_or_load, load, sb_off, lfsr_en, absorb, replace : std_logic;

  -- prevent SRL LUT configuration  for Xilinx
  attribute shreg_extract : string;
  attribute shreg_extract of sa : signal is "yes"; --"no";

begin

  load 		<= ctrl(load_idx );
  sb_off 	<= ctrl(sb_off_idx);
  lfsr_en 	<= ctrl(lfsr_en_idx);
  absorb 	<= ctrl(absorb_idx);
  replace	<= ctrl(replace_idx);
  -- tag crtl signal removed with new fsm
  	
  replace_or_load <= replace OR load;    -- stages with inputs D_9 , ... D_0

  -- outputs - nonlinears 2x(WGP, SB, SB)
  o_s36 <= sa(36);  o_s34 <= sa(34);   o_s27 <= sa(27); -- to left nonlinears
  o_s18 <= sa(18);  o_s15 <= sa(15);   o_s8  <= sa(8);  -- to right nonlinears

  -- output data
  output_data : for i in 0 to 36 generate  
    rate_if: if ratestages(i) /= D+1 generate        
      o_data(ratestages(i)) <= ab(i); 
    end generate  rate_if;
  end generate  output_data ;

  -- omega and feedback
  mo: entity work.lfsr_wage_omega port map (sa(0), omega);
  f <= sa(31) xor sa(30) xor sa(26) xor sa(24) xor sa(19) xor sa(13) xor sa(12) xor sa(8) xor sa(6) xor omega;
 
  -- from left nonlinears
  x(36) <= f      xor i_s36; 
  x(29) <= sa(30) xor i_s29;
  x(23) <= sa(24) xor i_s23;
 
  -- from right nonlinears
  x(18) <= sa(19) xor i_s18;
  x(10) <= sa(11) xor i_s10;
  x(4)  <= sa(5)  xor i_s4;

  i_data_9 <= i_data(8) when load = '1' else i_data(9);  -- loading ( 64 bits instead of 70)

  --****ABSORBING****-- 
  -- absorb: xor exception  
  ab(36) <= sa(36) xor i_data_9;
  ab(0)(0)     	      	<= sa(0)(0)  xor i_dom_sep(0);  
  ab(0)(1) 		          	<= sa(0)(1)  xor i_dom_sep(1);  
  ab(0)(2 to gf_dim - 1)   <= sa(0)(2 to gf_dim - 1);
   -- absorb: mux exception
  y(36) <= ab(36) when absorb = '1' else x(36);
  y(18) <= ab(18) when absorb = '1' else x(18);
  y(0)  <= ab(0)  when absorb = '1' else sa(1); -- dom sep  (needs absorb=1 during DEC as well)
  
  absorbgen : for i in 0 to 35 generate  
	  ab_27_17_8: for j in 0 to gf_dim -1  generate  
	    abxor_if: if absorbxor(i) = '1' generate  
        		ab(i)(j) <= sa(i)(j) xor i_data(ratestages(i))(j);
 	    end generate  abxor_if;
  		 abxortag_if: if absorbxortag(i) = '1' generate  
  	        ab(i)(j) <= sa(i)(j) xor ( not sb_off and i_data(ratestages(i))(j));      -- also TAG --
	    end generate  abxortag_if;
		 ab_if: if absorbplain(i) = '1' generate        
     		  y(i)(j) <= ab(i)(j) when absorb = '1' else sa(i+1)(j);
    	 end generate  ab_if;    
	  end generate ab_27_17_8;		    
  end generate  absorbgen;   
   
  --****LOADING/REPLACING ****--  
  -- replace muxes
  replacegen : for i in 0 to 35 generate  
    r_if: if replaceplain(i) = '1' generate        
      l(i) <= i_data(ratestages(i)) when replace = '1' else y(i);
    end generate  r_if;
    rl_if: if replaceload(i) = '1' generate        
      l(i) <= i_data(ratestages(i)) when replace_or_load = '1' else y(i);   
    end generate  rl_if;
  end generate  replacegen;

  l(36)(0)             <= i_data_9(0)             when replace_or_load='1'   else y(36)(0);             --RLmx36_bit0
  l(36)(1 to gf_dim-1) <= i_data_9(1 to gf_dim-1) when sb_off='1'            else y(36)(1 to gf_dim-1); --Lmx36_others
 
  -- SB muxes
   l(29) <= sa(30) when sb_off = '1' else x(29);      -- LOADING/TAG turn off SB
   l(23) <= sa(24) when sb_off = '1' else x(23);      -- LOADING/TAG turn off SB
   l(10) <= sa(11) when sb_off = '1' else x(10);      -- LOADING/TAG turn off SB
   l(4)  <= sa(5)  when sb_off = '1' else x(4);       -- LOADING/TAG turn off SB


  --****LFSR UPDATE****-- 

  -- plain shifts ( do NOT update regz on absorb = '1')
  lfsr_shiftplain : for i in 0 to 35 generate  
    lfsr_if: if shiftplain(i) = '1' generate        
    lfsr_step: process(clk) begin
      if rising_edge(clk) then
      if lfsr_en = '1' then        
          sa(i) <= sa(i+1);
      end if;  
      end if;
    end process;
   end generate  lfsr_if;
  end generate  lfsr_shiftplain;

  -- stages updated from load muxes
  cegen1: for i in 1 to 6 generate 
			lfsr_loadce1_shift: process(clk) begin
  				if rising_edge(clk) then
    				if lfsr_en = '1'  or absorb = '1' or replace = '1' then  
    				 	 if ce_ctrl(i) = '1'  then
			      		if  (pad_ctrl(i) = '1') then       
			         				sa(part1(i))(pb(i))		<= not  sa(part1(i))(pb(i));  --pad
			         	else 	  	sa(part1(i))(pb(i))		<= 		l(part1(i))(pb(i)); 		
			         	end if; 
         				sa(part2(i))(pb(i)  to 6)	<= l(part2(i))(pb(i)  to 6);  
						end if;      			
      			end if;
            end if;
      	end process;
  end generate  cegen1;

   -- stages updated from load muxes
  cegen2: for i in 1 to 6 generate 
		posgen: if pb(i)-1>=0 generate
			lfsr_loadce2_shift: process(clk) begin
  				if rising_edge(clk) then
    				if lfsr_en = '1'  or absorb = '1' or replace = '1' then  
     					 if ce_ctrl(i) = '1'  then
         				sa(part1(i))(0  to pb(i)-1)	<= l(part1(i))(0 to pb(i)-1);  
		      		end if;	
      			end if;
            end if;
      	end process;		
		end generate posgen;
  end generate  cegen2;		
  
  -- stages updated from load muxes: exceptions
  lfsr_load_shift: process(clk) begin
    if rising_edge(clk) then
      if lfsr_en = '1'  or absorb = '1' or replace = '1' then  
         if ce_ctrl(7) = '1' 	then 
							sa(36)(0 to 6)	<= l(36)(0 to 6);  
         				sa(35)(1 to 6)	<= l(35)(1 to 6);   
         end if; 
      	if ce_ctrl(0) = '1' 		then 
       		  if (pad_ctrl(0) = '1')  then 	 
         		   	sa(9)(1)			<= not sa(9)(1);    --pad
         	  else	sa(9)(1)			<= 	 l(9)(1);  
         	  end if;
         				sa(9)(0 to 0)	<= l(9)(0 to 0);     
 	     					sa(8)(0 to 6)  <= l(8)(0 to 6);   
   						sa(0)  			<= y(0);  		 -- dom sep 
   		end if; 
      end if;
    end if;
  end process;


  -- stages updated from SB muxes
  lfsr_sbmx_shift: process(clk) begin
    if rising_edge(clk) then
      if lfsr_en = '1' then        
	      sa(29) <= l(29);
	      sa(23) <= l(23);
	      sa(10) <= l(10);
	      sa(4)  <= l(4);
      end if;
    end if;
  end process;

end;

