-- This work is licensed under a Creative Commons
-- Attribution-NonCommercial-ShareAlike 4.0 International License.
-- http://creativecommons.org/licenses/by-nc-sa/4.0

-- 2020 modifications:  Mark D. Aagaard, Nusa Zidaric

-- Mark D. Aagaard
-- Riham AlTawy
-- Guang Gong
-- Kalikinkar Mandal
-- Raghvendra Rohit
-- Marat Sattarov
-- Nusa Zidaric
-- http://uwaterloo.ca/communications-security-lab/lwc/wage


-- This is a human-readable summary of (and not a substitute for) the license. 
-- You are free to:

--     Share:  copy and redistribute the material in any medium or format
--     Adapt:  remix, transform, and build upon the material

--     The licensor cannot revoke these freedoms as long as you follow
--     the license terms.

-- Under the following terms:

--     Attribution - You must give appropriate credit, provide a link to
--     the license, and indicate if changes were made. You may do so in
--     any reasonable manner, but not in any way that suggests the
--     licensor endorses you or your use.

--     NonCommercial - You may not use the material for commercial
--     purposes.

--     ShareAlike - If you remix, transform, or build upon the material,
--     you must distribute your contributions under the same license as
--     the original.

--     No additional restrictions - You may not apply legal terms or
--     technological measures that legally restrict others from doing
--     anything the license permits.


---------------------------------
--- TESTBENCH  for module WAGElfsr
--  test load correctness 

--  *passed*

---------------------------


LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.all;
USE STD.textio.all;
USE ieee.std_logic_textio.all;

use work.util_unsynth.all;
use work.util.all;
use work.tb_wage_unsynth.all;
use work.wage_pkg.all;    
use work.ctrl_pkg.all;
use work.tb_pkg.all;    

ENTITY tb_wage IS
  generic ( period : real   -- := 10.0
          ; ED     : std_logic_vector(0 to 1) := "11" -- Encryption | Decryption               
          ; S      : std_logic_vector(0 to 1) := "00" -- stall flag: 00 - none, 01 - stall enc, 10 - stall dec, 11 stall other   
         );
END tb_wage; 

ARCHITECTURE main OF tb_wage IS

  file simo_file      : text open write_mode is simo_file_path;
  file stim_file      : text;

  constant clk_period   : time  := period * 1 ns;
  constant hold         : time  := clk_period / 5;
  constant after_reset  : time  := 0 * clk_period;  -- delay between reset and loading
  constant load_delay   : time  := 0 * clk_period;  -- delay between K, N data
  constant init_delay   : time  := 0 * clk_period;  -- delay between load permutation and initialization
  constant procad_delay : time  := 0 * clk_period;  -- AD data delay
  constant stall_short_delay  : time  := 3   * clk_period;  -- stall short delay
  constant stall_long_delay   : time  := 300 * clk_period;  -- stall long delay
  constant stall_period       : time  := 5   * clk_period;   
  constant enc_delay    : time  := 10 * clk_period;  -- encryption data delay
  constant dec_delay    : time  := 10 * clk_period;  -- decryption data delay
  constant mode_delay   : time  := 50 * clk_period;  -- delay between two modes
 
 
 
  signal clk, reset : std_logic := '0'; 
  signal stall : boolean := false; 
  
  signal tag_data : std_logic_vector(0 to 127);
  signal tag_valid  : std_logic := '0';
  type tb_state_ty is (tbLoad, tbInit, tbProcAD, tbEncrypt, tbDecrypt, tbFinal, tbTag, tbNull);
  signal tb_state : tb_state_ty;   

  signal pre_data, post_data    : word;
  signal pre_valid, pre_ready   : std_logic;
  signal post_valid, post_ready : std_logic;
  signal pre_dom_sep     : dom_sep_ty;
  signal pre_dom_sep_buf : dom_sep_ty;
  signal pre_mode        : mode_ty;
  signal pre_mode_buf    : mode_ty;
  signal pre_padding     : full_pad_ty;
  signal bits_unpadded   : integer;
  
  type state_array_temp is array (8 downto 0) of std_logic_vector(31 downto 0);

BEGIN  

  pre_dom_sep_proc : process(pre_dom_sep_buf)
  begin
    pre_dom_sep <= pre_dom_sep_buf after hold;
  end process;

  pre_mode_proc : process(pre_mode_buf)
  begin
    pre_mode <= pre_mode_buf after hold;
  end process;

  post_reading_proc : process -- reading the tb/sim output , writing the simo file 
    variable msg    : line;
    variable done, skip_write  : boolean;
    --alias post_data_alias is << signal .wage_dp_tb.uut.post_data : data_io >>;
    variable post_data_var : std_logic_vector(0 to 63);
  begin

    skip_write := false;
    done := false;
    while not(done) loop
      wait until rising_edge(clk);
      done := (post_valid = '1' or tag_valid = '1' or (pre_valid = '1' and pre_ready='1'));
    end loop;

    post_data_var := post_data;

	    if tb_state = tbEncrypt then
			 if 	bits_unpadded >= 0  then    
			      write(  msg, "C " & to_hex_string_normal( post_data( 0 to bits_unpadded-1 ) ));     
			 else
	       		skip_write := true;
	      end if;
	    elsif tb_state = tbDecrypt then
		    if 	bits_unpadded >= 0 then        
	     			write(  msg, "P " & to_hex_string_normal( post_data( 0 to bits_unpadded-1 ) ));
			 else 
	       		skip_write := true;
	       end if;     			
	     			
	    elsif tb_state = tbTag then
	      if tag_valid = '1' then
	        write(  msg, "T " & to_hex_string_normal( tag_data ));
	      else
	        skip_write := true;
	      end if;
	    else
	      skip_write := true;
	    end if;
	    
    if not(skip_write) then writeline( simo_file, msg); end if;

  end process;
 
  post_stalling_proc : process 

  begin
    	
      if S = "01" then
      	if   tb_state = tbEncrypt then  
	 		 	stall <= true;
				wait for stall_long_delay; 
	 		end if;		
	 		stall <= false;
	 		wait for stall_period; 
	 	elsif S = "10" then 
      	if   tb_state = tbDecrypt then  
	 		 	stall <= true;
				wait for stall_long_delay; 
	 		end if;	
	 		stall <= false;
	 		wait for stall_period; 	
	 	elsif S = "11" then 
      	if tb_state = tbInit or tb_state = tbProcAD or tb_state = tbFinal or  tb_state = tbTag then  
	 		 	stall <= true;
				wait for stall_long_delay; 
	 		end if;	
	 		stall <= false;
	 		wait for stall_period;  	
	 	
 		else
 		   stall <= false; 	 
 		   wait for stall_period; 
  		end if;
		  
  end process;
  
  post_ready <= '0' when stall else '1';
  clk <= not(clk) after clk_period / 2;


   -- Instantiate the Unit Under Test (UUT)
  uut: entity work.wage port map 
    (  clk       => clk
    ,  reset     => reset
    ,  pre_mode    => pre_mode
    ,  pre_dom_sep => pre_dom_sep
    ,  pre_padding => pre_padding
    ,  pre_data    => pre_data
    ,  pre_valid   => pre_valid
    ,  post_ready   => post_ready
    ,  pre_ready    => pre_ready
    ,  post_data    => post_data
    ,  post_valid   => post_valid
    );

  stim_proc: process
 
    variable target_stim		: stim_ty;
    variable key_rd, nonce_rd : std_logic_vector(0 to 127);
    variable k0, k1, data     : std_logic_vector(0 to 63);
    variable tag_vector       : data_io_vector(0 to 8); -- actual tag
    variable kn_load_vector   : data_io_vector(0 to 8);

  begin  

    if ED(0) = '1' then
      ------------------------------------------------------------------------------
      ---------------------------------- Encryption --------------------------------
      ------------------------------------------------------------------------------
	 
      tag_data <= (others => 'X');
      pre_valid  <= '0';
      pre_padding  <= (others => '0');
 
      pre_mode_buf <= '0'; -- new FSM requires this stable 
      tb_state <= tbNull;

      -------------------------------- Reset -------------------------------

      wait for clk_period;
      drive_reset( clk, reset, hold );
      wait until rising_edge(clk);
  
      -------------------------------- Load --------------------------------
      report( "LOADING PHASE" );
      tb_state <= tbLoad;

      pre_dom_sep_buf <= "00";

      read_key_nonce( key_rd, nonce_rd, stim_file_path );
      kn_load_vector := kn_to_data_io_vector (key_rd, nonce_rd);
      k0 := key_rd(0 to 63);
      k1 := key_rd(64 to 127);

      for i in 0 to 8 loop
        data := data_io_to_m(kn_load_vector(i));
        drive_data( clk, data, hold, pre_ready, pre_data, pre_valid );
        wait until rising_edge(clk);
      end loop;
 
      -------------------------------- LoadPerm ------------------------------
      report( "PERMUTATION AFTER LOAD" );   
      wait until rising_edge(clk); 
      wait for init_delay;

      ---------------------------------- Init --------------------------------
      report( "INITIALIZATION PHASE" );
      tb_state <= tbInit;
		 
      drive_data( clk, k0, hold, pre_ready, pre_data, pre_valid );
      wait until rising_edge(clk);
      drive_data( clk, k1, hold, pre_ready, pre_data, pre_valid );
      wait until rising_edge(clk);

      ---------------------------------- ProcAD --------------------------------
      report( "AD PROCESSING PHASE" );
      tb_state <= tbProcAD;

      pre_dom_sep_buf <= "01"; 
	
		target_stim := AD_STIM;
      drive_all(target_stim, stim_file_path, hold, procad_delay, clk, pre_ready, pre_data, pre_valid, pre_padding,  bits_unpadded);
 
      ---------------------------------- Encrypt ------------------------------
      report( "ENCRYPTION PHASE" );
      tb_state <= tbEncrypt;

      pre_dom_sep_buf <= "10";

      
		target_stim := PLAINTEXT_STIM;
      drive_all(target_stim, stim_file_path, hold, enc_delay, clk, pre_ready, pre_data, pre_valid, pre_padding,  bits_unpadded);

      ---------------------------------- Final --------------------------------
      report( "FINALIZATION PHASE" );
      tb_state <= tbFinal;
  
      pre_dom_sep_buf <= "00";

		target_stim := KEY_STIM;
      drive_data( clk, k0, hold, pre_ready, pre_data, pre_valid );
      wait until rising_edge(clk);

      drive_data( clk, k1, hold, pre_ready, pre_data, pre_valid );
      wait until rising_edge(clk);

      ---------------------------------- Tag ----------------------------------
      report( "TAG PHASE" );
      tb_state <= tbTag;
   
      for i in 0 to 8 loop 
        	wait until post_valid='1' and post_ready ='1' and rising_edge(clk);        	
      	tag_vector(i) := m_to_data_io(post_data);
        	report(    "tag vector was read for i=" &integer'image(i) &" : " & to_hex_string_normal( post_data )       );
      end loop;

      tag_data <= data_io_vector_to_tag(tag_vector);
      tag_valid <= '1';

      wait until rising_edge(clk); wait for hold;
      report("tag is: " & to_hex_string_normal( tag_data ) );
      tag_valid <= '0';

      wait for mode_delay;

    end if;

    if ED(1) = '1' then
      ------------------------------------------------------------------------------
      ---------------------------------- Decryption --------------------------------
      ------------------------------------------------------------------------------
 
      tag_data <= (others => 'X');
      pre_valid    <= '0'; 
      pre_padding  <= (others => '0');
    
      tb_state <= tbNull;
      pre_mode_buf <= '1';
      -------------------------------- Reset -------------------------------

      wait for clk_period;
      drive_reset( clk, reset, hold );
      wait until rising_edge(clk);
  
      -------------------------------- Load --------------------------------
      report( "LOADING PHASE" );
      tb_state <= tbLoad;

      pre_dom_sep_buf <= "00";

      read_key_nonce( key_rd, nonce_rd, stim_file_path );
      kn_load_vector := kn_to_data_io_vector (key_rd, nonce_rd);
      k0 := key_rd(0 to 63);
      k1 := key_rd(64 to 127);

      for i in 0 to 8 loop
        data := data_io_to_m(kn_load_vector(i));
        drive_data( clk, data, hold, pre_ready, pre_data, pre_valid );
        wait until rising_edge(clk);
      end loop;
     
      -------------------------------- LoadPerm ------------------------------
      report( "PERMUTATION AFTER LOAD" );   
      wait until rising_edge(clk); 
      wait for init_delay;

      ---------------------------------- Init --------------------------------
      report( "INITIALIZATION PHASE" );
      tb_state <= tbInit;

      drive_data( clk, k0, hold, pre_ready, pre_data, pre_valid );
      wait until rising_edge(clk);
      drive_data( clk, k1, hold, pre_ready, pre_data, pre_valid );
      wait until rising_edge(clk);

      ---------------------------------- ProcAD --------------------------------
      report( "AD PROCESSING PHASE" );
      tb_state <= tbProcAD;

      pre_dom_sep_buf <= "01";

		target_stim := AD_STIM;
      drive_all(target_stim, stim_file_path, hold, procad_delay,  clk,  pre_ready, pre_data, pre_valid, pre_padding,   bits_unpadded);
 
      ---------------------------------- Decrypt ------------------------------
      report( "DECRYPTION PHASE" );
      tb_state <= tbDecrypt;

      pre_dom_sep_buf <= "10";

		target_stim := CIPHERTEXT_STIM; 
      drive_all(target_stim, stim_file_path, hold, enc_delay,  clk, pre_ready,  pre_data, pre_valid, pre_padding,  bits_unpadded);
 
      ---------------------------------- Final --------------------------------
      report( "FINALIZATION PHASE" );
      tb_state <= tbFinal;
  
      pre_dom_sep_buf <= "00";

		target_stim := KEY_STIM;
      drive_data( clk, k0, hold, pre_ready, pre_data, pre_valid );
      wait until rising_edge(clk);

      drive_data( clk, k1, hold, pre_ready, pre_data, pre_valid );
      wait until rising_edge(clk);

      ---------------------------------- Tag ----------------------------------
      report( "TAG PHASE" );
      tb_state <= tbTag;
   
      for i in 0 to 8 loop
            wait until post_valid='1' and post_ready ='1' and rising_edge(clk);
	   		tag_vector(i) := m_to_data_io(post_data);
        		report(    "tag vector was read for i=" &integer'image(i) &" : " & to_hex_string_normal( post_data )       );
      end loop;

      tag_data <= data_io_vector_to_tag(tag_vector);
      tag_valid <= '1';

      wait until rising_edge(clk); wait for hold;
      report("tag is: " & to_hex_string_normal( tag_data ) );
      tag_valid <= '0';

      wait for mode_delay;
  
    end if;

      assert false
      report ("SIMULATION IS FINISHED")
      severity failure;

  end process;

end;
