-- This work is licensed under a Creative Commons
-- Attribution-NonCommercial-ShareAlike 4.0 International License.
-- http://creativecommons.org/licenses/by-nc-sa/4.0

-- 2020 modifications:  Mark D. Aagaard, Nusa Zidaric

-- Mark D. Aagaard
-- Riham AlTawy
-- Guang Gong
-- Kalikinkar Mandal
-- Raghvendra Rohit
-- Marat Sattarov
-- Nusa Zidaric
-- http://uwaterloo.ca/communications-security-lab/lwc/wage


-- This is a human-readable summary of (and not a substitute for) the license. 
-- You are free to:

--     Share:  copy and redistribute the material in any medium or format
--     Adapt:  remix, transform, and build upon the material

--     The licensor cannot revoke these freedoms as long as you follow
--     the license terms.

-- Under the following terms:

--     Attribution - You must give appropriate credit, provide a link to
--     the license, and indicate if changes were made. You may do so in
--     any reasonable manner, but not in any way that suggests the
--     licensor endorses you or your use.

--     NonCommercial - You may not use the material for commercial
--     purposes.

--     ShareAlike - If you remix, transform, or build upon the material,
--     you must distribute your contributions under the same license as
--     the original.

--     No additional restrictions - You may not apply legal terms or
--     technological measures that legally restrict others from doing
--     anything the license permits.


use std.textio.all;
library ieee;
use ieee.math_real.all;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_textio.all;

use work.util_unsynth.all;
use work.wage_pkg.all;

package tb_wage_unsynth is

  constant max_data_sz     : integer := 320;
  -- ******************************************** --
  type stim_ty is ( KEY_STIM, NONCE_STIM, INPUT_STIM, OUTPUT_STIM, AD_STIM, PLAINTEXT_STIM, CIPHERTEXT_STIM, EOF_STIM, NULL_STIM );


  function to_hex_string_normal( v : std_logic_vector ) return string;
  -- ******************************************** --
  function is_stim( c : character ) return boolean;
  function chr_to_stim( chr : character ) return stim_ty;
  -- ******************************************** --
  
 

  procedure read_next_stim_data_line_normal
    ( file f   : text;
      stim     : out stim_ty;
      data     : out std_logic_vector;
      num_bits : out natural );
      
  procedure hread_dir_normal
    ( ln       : inout line;
      result   : out   std_logic_vector;
      num_bits : out   natural
    );

  function m_to_data_io	(m : std_logic_vector( 0 to 63 ) ) return data_io;
  function data_io_vector_to_tag ( s : data_io_vector( 0 to 8 ) ) return std_logic_vector;
  function kn_to_data_io_vector (key_rd, nonce_rd : std_logic_vector( 0 to 127 ) ) return data_io_vector;
 
  function pad ( v : std_logic_vector ) return std_logic_vector;



  function data_io_to_m (s : data_io ) return std_logic_vector;

  procedure drive_reset
    ( signal   clk   : in std_logic
    ; signal   reset : out std_logic
    ; constant hold  : in time
    );

  procedure read_key_nonce
    ( key_rd       : out std_logic_vector
    ; nonce_rd     : out std_logic_vector
    ; stim_file_path : in string
    );


  procedure drive_data
    ( signal   clk     : in std_logic
    ; variable data    : in std_logic_vector
    ; constant hold    : in time
    ; signal   pre_ready  : in std_logic
    ; signal   pre_data   : out std_logic_vector
    ; signal   pre_valid  : out std_logic

    );

  procedure drive_all
    ( constant target_stim : in stim_ty
    ; stim_file_path       : in string
    ; constant hold        : in time
    ; constant delay       : in time
 
    ; signal clk           : in std_logic
    ; signal pre_ready       : in std_logic 
 
    ; signal pre_data        : out std_logic_vector
    ; signal pre_valid       : out std_logic
    ; signal pre_padding     : out full_pad_ty
    ; signal bits_unpadded   : out integer
    );

end package;  


package body tb_wage_unsynth is

  procedure drive_reset
    ( signal   clk   : in std_logic
    ; signal   reset : out std_logic
    ; constant hold  : in time
    )
  is
  begin

    wait until rising_edge(clk);
    wait for hold;
    reset <= '1';                                                
    wait until rising_edge(clk);
    wait for hold;
    reset <= '0';

  end procedure;


  procedure read_key_nonce
    ( key_rd          : out std_logic_vector
    ; nonce_rd        : out std_logic_vector
    ; stim_file_path    : in string
    )
  is
    variable stim_rd   : stim_ty;
    file     stim_file : text;
    variable num_bits  : natural;
  begin

    file_close(stim_file);
    file_open(stim_file, stim_file_path, read_mode);

    stim_rd := NULL_STIM;
    while stim_rd /= KEY_STIM and stim_rd /= EOF_STIM loop
      read_next_stim_data_line_normal( stim_file, stim_rd, key_rd, num_bits );
    end loop;

    if stim_rd = KEY_STIM then
      if num_bits /= 128 then
        file_close(stim_file);
        assert false
        report( "Key length = " & natural'image(num_bits) & " bits, must be 128 bits")
        severity failure;
      else
        report( "Key length = " & natural'image(num_bits) & " bits");
      end if;
    else
      file_close(stim_file);
      assert false
      report ("Key isn't found in file " & stim_file_path)
      severity failure;
    end if;

    file_close(stim_file);
    file_open(stim_file, stim_file_path, read_mode);

    stim_rd := NULL_STIM;
    while stim_rd /= NONCE_STIM and stim_rd /= EOF_STIM loop
      read_next_stim_data_line_normal( stim_file, stim_rd, nonce_rd, num_bits );
    end loop;

    if stim_rd = NONCE_STIM then
      if num_bits /= 128 then
        file_close(stim_file);
        assert false
        report( "Nonce length = " & natural'image(num_bits) & " bits, must be 128 bits")
        severity failure;
      else
        report( "Nonce length = " & natural'image(num_bits) & " bits");
      end if;
    else
      file_close(stim_file);
      assert false
      report ("Nonce isn't found in file " & stim_file_path)
      severity failure;
    end if;

    file_close(stim_file);

  end procedure;


  procedure drive_data
    ( signal   clk      : in std_logic
    ; variable data     : in std_logic_vector
    ; constant hold     : in time
    ; signal   pre_ready  : in std_logic
    ; signal   pre_data   : out std_logic_vector
    ; signal   pre_valid  : out std_logic

    )
  is
  begin

    wait for hold;
    pre_data   <= data;
    pre_valid  <= '1';
    wait until pre_ready = '1' and rising_edge(clk);	
--    wait for hold;  --   causes transition to pFinal instead pInit 
    pre_valid  <= '0';

  end procedure;
 

 
  procedure drive_all
    ( constant target_stim : in stim_ty
    ; stim_file_path       : in string
    ; constant hold        : in time
    ; constant delay       : in time
 
    ; signal clk           : in std_logic
    ; signal pre_ready       : in std_logic 
 
    ; signal pre_data        : out std_logic_vector
    ; signal pre_valid       : out std_logic
    ; signal pre_padding     : out full_pad_ty
    ; signal bits_unpadded   : out integer  -- needed for writing the stim output ( cut Ctxt/Ptxt to correct length) - change to integer
    )
  is
    variable stim_rd  : stim_ty; --returned by the read_next_stim_data_line_normal
    variable num_bits, last_num_bits : natural;
    variable data: std_logic_vector(0 to 63);
    variable data128 : std_logic_vector(0 to 127);
    file     stim_file : text;

  begin
    pre_padding <= (others => '0');
    bits_unpadded  <= 64;    
	 last_num_bits := 0;
	 	 
    file_close(stim_file);
    file_open(stim_file, stim_file_path, read_mode);

    stim_rd := NULL_STIM;
    while stim_rd /= target_stim and stim_rd /= EOF_STIM loop
      read_next_stim_data_line_normal( stim_file, stim_rd, data128, num_bits );
    end loop;

    while stim_rd /= EOF_STIM loop    

      report( stim_ty'image(stim_rd)  & " - data length = " & natural'image(num_bits) & " bits");
		
		last_num_bits := num_bits; 	-- will trigger full pad if 64 or 128
		
		if num_bits = 0 then																			-- empty, full plock pad if ptxt/ctxt 
		  	 if  stim_rd = PLAINTEXT_STIM	 or   stim_rd = CIPHERTEXT_STIM then	  	
              data := ( 0 => '1' , others => '0');											-- FULL  PAD  NOW    
		  		  bits_unpadded  <= 0;   -- exception
				  pre_padding <= ( 3 => '1', others => '0');
 		        drive_data( clk, data, hold, pre_ready, pre_data, pre_valid ); -- drive full pad 
	     	wait until  rising_edge(clk);		        
		          
          end if;  
			 -- if AD_STIM dont drive anything!!! will go look for ptxt/ctxt instead
			 
      elsif num_bits < 64 then 																	-- one word padded
   	   data := data128(0 to 63); 
   	   data := pad( data ); 
 			bits_unpadded  <= num_bits;     
 			pre_padding(2 downto 0) <= std_logic_vector(to_unsigned( num_bits/8 , 3) );
 			pre_padding(3) <= '0';
  			-- drive first 64 bits:
     	  	drive_data( clk, data, hold, pre_ready, pre_data, pre_valid );
  		  	wait until  rising_edge(clk);		  
	
		elsif num_bits = 64 then 																	-- one word as is, second word = full pad 
	 		data := data128(0 to 63); 
 			bits_unpadded  <= num_bits;    	
 			pre_padding <= (others => '0');								
   		-- drive first 64 bits:       	
       	drive_data( clk, data, hold, pre_ready, pre_data, pre_valid );
   	   wait until  rising_edge(clk);		  
  	
      elsif num_bits < 128 then 																	-- two words: first word as is, second word padded
   		data :=  data128(0 to 63); 
 			bits_unpadded  <= 64;   
 			pre_padding <= (others => '0');
  			-- drive first 64 bits: 		
	  	  	drive_data( clk, data, hold, pre_ready, pre_data, pre_valid );
			wait until  rising_edge(clk);		      
            
   	   data := data128(64 to 127); 
   	   data := pad( data ); 
 			bits_unpadded  <= num_bits - 64 ;     
 	 		pre_padding(2 downto 0) <= std_logic_vector(to_unsigned( num_bits/8 , 3) );
 			pre_padding(3) <= '0';
  			-- drive second 64 bits:
     	  	drive_data( clk, data, hold, pre_ready, pre_data, pre_valid );
			wait until  rising_edge(clk);		  

      elsif num_bits = 128 then 																	-- three words: first word as is, second word as is, third word = full pad
   		data :=  data128(0 to 63); 
 			bits_unpadded  <= 64;   
 			pre_padding <= (others => '0');
 		 	-- drive first 64 bits: 		
        	drive_data( clk, data, hold, pre_ready, pre_data, pre_valid );
	   	wait until  rising_edge(clk);		       
      
   	   data :=   data128(64 to 127); 
 			bits_unpadded  <= num_bits - 64 ; 
 			pre_padding(2 downto 0) <= std_logic_vector(to_unsigned( num_bits/8 , 3) );
 			pre_padding(3) <= '0';
 			-- drive second 64 bits:
       	drive_data( clk, data, hold, pre_ready, pre_data, pre_valid );
	      wait until  rising_edge(clk);		      
  	
      end if;


--- when stim data length > 128 will read it in the next line:
      stim_rd := NULL_STIM;
      while stim_rd /= target_stim and stim_rd /= EOF_STIM loop
        read_next_stim_data_line_normal( stim_file, stim_rd, data128, num_bits );
      end loop;

    end loop;

	 if last_num_bits = 64 or last_num_bits = 128 then 
         data := ( 0 => '1' , others => '0'); 		  										-- FULL  PAD 
  			bits_unpadded  <= -1;    
		   pre_padding <= ( 3 => '1', others => '0');
 			-- drive full pad 
     	  	drive_data( clk, data, hold, pre_ready, pre_data, pre_valid );
      	wait until  rising_edge(clk);		  

	 end if;

    file_close(stim_file);
  
  end procedure;

  function data_io_to_m (s : data_io ) return std_logic_vector is
    variable z : std_logic_vector (0 to 69);
    variable z2 : std_logic_vector (0 to 63);
  begin

    for i in 0 to 9 loop
      for j in 0 to 6 loop
         z(i*7+j) := s(i)(j);
      end loop;
    end loop;

  z2 := z(0 to 63);
  return z2;

  end function;

  -- ******************************************** --
  -- ******************************************** --
  function m_to_data_io	(m : std_logic_vector( 0 to 63 ) ) return data_io is
    variable z	: data_io;
  begin

    for i in 0 to 8 loop
      z(i) := m(i*7 to i*7+6);
    end loop;

    z(9) := m(63) & "000000";
    return z;

  end function;
  -- ******************************************** --
  -- ******************************************** --
  -- tag extract
  function data_io_vector_to_tag ( s : data_io_vector(0 to 8) ) return std_logic_vector is
    variable z	: std_logic_vector( 0 to 127 );
  begin
    
    for i in 0 to 8 loop
      z(2*i*7 to 2*i*7 + 6) := s(i)(6);		-- T0 to T16
    end loop;

    for i in 0 to 7 loop
      z((2*i+1)*7 to (2*i+1)*7 + 6) := s(i)(1);	-- T1 to T15
    end loop;

    z(17*7 to 17*7 + 6) := s(1)(3);		-- T17
    z(126 to 127)	:= s(2)(3)(2 to 3);	-- T18

    return z;

  end function;
  -- ******************************************** --
  -- ******************************************** --
  -- key/nonce loading sequence 
  function kn_to_data_io_vector (key_rd, nonce_rd : std_logic_vector( 0 to 127 ) )
  return data_io_vector is
    variable z		: data_io_vector(0 to 8);
    variable n_vector	: gf_vector(0 to 17);
    variable k_vector	: gf_vector(0 to 17);
    variable m_vector	: gf_vector(0 to 17);
    variable k18	: gf_elem;
  begin

    n_vector := (others => (others => '0') );
    k_vector := (others => (others => '0') );

    for i in 0 to 8 loop
      for j in 0 to gf_dim-1 loop
	n_vector(i)(j) := nonce_rd(7*i+j);
	k_vector(i)(j) := key_rd(7*i+j);
      end loop;
    end loop;

    for i in 9 to 17 loop
      for j in 0 to gf_dim-1 loop
	n_vector(i)(j) := nonce_rd(7*i+1+j);
	k_vector(i)(j) := key_rd(7*i+1+j);
      end loop;
    end loop;

    k18 := key_rd(63) & key_rd(127) & nonce_rd(63) & nonce_rd(127) & "000";
    z := (others => (others => (others => '0')));

    for i in 0 to 8 loop
      z(i)(0) := k_vector(2*i);
      z(i)(5) := k_vector(1+2*i);
      z(i)(8) := n_vector(2*i);
    end loop;

    for i in 1 to 8 loop
      z(i)(3) := n_vector(2*i-1);
    end loop;

    z(7)(4) := n_vector(15);
    z(8)(4) := k18;

    z(7)(3) := n_vector(13);
    z(8)(3) := n_vector(17);

    return z;

  end function;

  -- ******************************************** --
  -- ******************************************** --
  -- pad with 10* rule
  function pad ( v : std_logic_vector ) return std_logic_vector is
    variable z : std_logic_vector(v'low to v'high);
  begin
    report( "PADDING for v = " & to_string(v) );
    for i in v'high downto v'low + 1 loop
      if v(i) /= '1' and v(i) /= '0' then
			if v(i-1) = '1' or v(i-1) = '0' then
	 		 z(i) := '1';
			else
	  		 z(i) := '0';
		end if;
      else
			z(i) := v(i);
      end if;
    end loop;
    z(0) := v(0);
    return(z);
  end function;
  -- ******************************************** --
  -- ******************************************** --
  function to_hex_string_normal( v : std_logic_vector ) return string is
  begin
    if v'length = 0 then 
      return "";
    elsif v'length <= 4 then
      return to_hex_char( rev( v ) ) & "";				 
    else
      if v'ascending then

        return to_hex_char( rev( v( v'low to v'low+3)) )		 
               & to_hex_string_normal( v(v'low+4 to v'high) );
      else

        return to_hex_char( rev( v( v'high downto v'high-3)) )		 
               & to_hex_string_normal( v(v'high-4 downto v'low) );
      end if;
    end if;
  end function;
  -- ******************************************** --
  -- ******************************************** --
  function is_stim( c : character ) return boolean is
  begin
    return c = 'K' or c = 'N' or c = 'I' or c = 'O' or c = 'A' or c = 'P' or c = 'C';
  end function;
  -- ******************************************** --
  -- ******************************************** --
  function chr_to_stim( chr : character ) return stim_ty is
  begin
    case chr is
      when 'K' => return KEY_STIM;
      when 'N' => return NONCE_STIM;
      when 'I' => return INPUT_STIM;
      when 'O' => return OUTPUT_STIM;
      when 'A' => return AD_STIM;
      when 'P' => return PLAINTEXT_STIM;
      when 'C' => return CIPHERTEXT_STIM;
      when others =>
        assert false
          report ("chr_to_stim: not a stim: "& character'image( chr ))
          severity error;
        return EOF_STIM;
    end case;
  end function;
  -- ******************************************** --
  -- ******************************************** --
   procedure read_next_stim_data_line_normal
    ( file f   : text;
      stim     : out stim_ty;
      data     : out std_logic_vector;
      num_bits : out natural )
  is
    variable found
           , ok            : boolean;
    variable input_line    : line;
    variable chr           : character;
    variable tmp_stim      : stim_ty;
    variable tmp_data      : std_logic_vector( data'range );
    variable tmp_num_bits  : natural;
  begin
    -- read lines until hit one where first non-blank character is a stim charcter 
    found := false;
    while not found and not endfile( f ) loop
      readline( f, input_line ); 
      read( input_line, chr, good => ok );
      -- skip blanks and comments
      found := ok and is_stim(chr);
    end loop;
    if not found then
      tmp_stim  := EOF_STIM;
      tmp_data := (others => 'U');
    else
      tmp_stim  := chr_to_stim( chr );
      hread_dir_normal( input_line, tmp_data, tmp_num_bits );      
      
    end if;
    if tmp_num_bits > 0 then 
      report( "read_next_stim_data: --> "& stim_ty'image(tmp_stim) &" "&
              to_string( firstn( tmp_data, tmp_num_bits ) ) );
    else
      report( "read_next_stim_data: --> "& stim_ty'image(tmp_stim) &" 0 bits");
    end if;
    stim     := tmp_stim;
    data     := tmp_data;
    num_bits := tmp_num_bits;
  end procedure;
  -- ******************************************** --
  -- ******************************************** --
 procedure hread_dir_normal
    ( ln       : inout line;
      result   : out   std_logic_vector;
      num_bits : out   natural
    )
  is
    variable ok           : boolean;
    variable chr          : character;
    variable str          : string(1 to result'length/4);
    variable tmp_result   : std_logic_vector(0 to result'length-1);
    variable num_chars,
             tmp_num_bits : natural := 0;
  begin
    
    if result'length mod 4 /= 0 then
      assert false
        report 
          "hread_dir error: trying to read vector " &
             "with non multiple of 4 length";
      return;
    end if;

    loop                                    -- skip white space
      read(ln, chr, ok);
      exit when not ok or ((chr /= ' ') and (chr /= CR) and (chr /= HT));
    end loop;
    if not ok then
      assert false
        report ("hread_dir: end of line")         severity warning;
      return;
    else
     -- report ("hread_dir: char0="& chr );
    end if;

    num_chars        := 1;
    str( num_chars ) := chr;
    
    while ok loop
      read(ln, chr, ok);
      if ok then 
        num_chars        := num_chars + 1;
        str( num_chars ) := chr;
      end if;
    end loop;

    tmp_num_bits := num_chars * 4;
    tmp_result   := (others => 'U');
    if result'ascending then
      tmp_result( 0 to tmp_num_bits - 1) :=
        hex_string_to_dn_vector( str( 1 to num_chars ) );   -- <––– changed from up to dn (change 1/2)
                                                            -- HEX(0) should be treated MSB, HEX(3) should be treated as LSB
                                                            -- for correct interpretation of HEX
    else 
      tmp_result( 0 to tmp_num_bits - 1) :=
        hex_string_to_up_vector( str( 1 to num_chars ) );   -- <––– changed from up to dn (change 2/2)
                                                            -- HEX(0) should be treated MSB, HEX(3) should be treated as LSB
                                                            -- for correct interpretation of HEX
    end if;
    
    -- report( "hread_dir:  " & str &" -->"&
    --        to_string( tmp_result( 0 to tmp_num_bits - 1 ) ) );
    
    result   := tmp_result;
    num_bits := tmp_num_bits;
  end procedure; 

end package body;
